import json
import pytest
import numpy as np

from memo import memfile


def confirm_file_contents(fpath, data):
    with open(fpath, "r") as f:
        data_read = [json.loads(j) for j in f.read().split("\n") if j != ""]
    for i, d in enumerate(data_read):
        assert data[i] == d


@pytest.mark.parametrize("kw", [{"a": 1, "b": 1}, {"a": 3, "b": 4}])
def test_base_values(kw, tmp_path):
    filepath = f"{tmp_path}/file.jsonl"

    @memfile(filepath=filepath)
    def count_values(**kwargs):
        return {"sum": sum(kwargs.values())}

    count_values(**kw)

    confirm_file_contents(filepath, [{**kw, "sum": kw["a"] + kw["b"]}])


def test_base_multiple_calls(tmp_path):
    filepath = f"{tmp_path}/file.jsonl"

    @memfile(filepath=filepath)
    def count_values(**kwargs):
        return {"sum": sum(kwargs.values())}

    for i in range(1, 5):
        count_values(a=1)
        confirm_file_contents(filepath, [{"a": 1, "sum": 1}] * i)


def test_also_works_with_numpy1(tmp_path):
    filepath = f"{tmp_path}/file.jsonl"

    @memfile(filepath=filepath)
    def count_values(**kwargs):
        return {"sum": sum(kwargs.values())}

    for i in range(1, 5):
        count_values(a=np.array([1]))
        confirm_file_contents(filepath, [{"a": [1], "sum": [1]}] * i)


def test_also_works_with_numpy2(tmp_path):
    filepath = f"{tmp_path}/file.jsonl"

    @memfile(filepath=filepath)
    def count_values(**kwargs):
        return {"sum": sum(kwargs.values())}

    count_values(a=np.array([1])[0])
    confirm_file_contents(filepath, [{"a": 1, "sum": 1}])



import pytest
from memo import memlist


@pytest.mark.parametrize("kw", [{"a": 1, "b": 1}, {"a": 3, "b": 4}])
def test_base_values(kw):
    data = []

    @memlist(data=data)
    def count_values(**kwargs):
        return {"sum": sum(kwargs.values())}

    count_values(**kw)

    assert len(data) == 1
    assert data[0]["a"] == kw["a"]
    assert data[0]["b"] == kw["b"]
    assert data[0]["sum"] == kw["a"] + kw["b"]


@pytest.mark.parametrize("kw", [{"a": 1}, {"a": 3, "b": 4}, {"a": 3, "b": 4, "c": 5}])
def test_keys_included(kw):
    data = []

    @memlist(data=data)
    def count_values(**kwargs):
        return {"sum": sum(kwargs.values())}

    count_values(**kw)

    assert len(data) == 1
    assert all([k in data[0].keys() for k in kw.keys()])


def test_base_multiple_calls():
    data = []

    @memlist(data=data)
    def count_values(**kwargs):
        return {"sum": sum(kwargs.values())}

    for i in range(1, 5):
        count_values(a=1)
        assert len(data) == i


def test_base_args():
    data = []

    @memlist(data=data)
    def count_values(a, b, **kwargs):
        return {"sum": sum(kwargs.values())}

    count_values(1, b=2, c=1)
    assert len(data) == 1
    assert data[0]["b"] == 2
    assert data[0]["c"] == 1


def test_base_args_included():
    data = []

    @memlist(data=data)
    def count_values(a, b, **kwargs):
        return {"sum": sum(kwargs.values())}

    count_values(a=1, b=2, c=1)
    assert len(data) == 1
    assert data[0]["a"] == 1
    assert data[0]["b"] == 2
    assert data[0]["c"] == 1



from functools import wraps

import httpx


def memweb(url: str):
    """
    Remembers input/output of a function by sending it over http to an endpoint.

    Important:
        Note that this decorator requires an extra dependeny. Ensure it is installed
        properly by running either;

        ```
        python -m pip install "memo[web]"
        ```

        You can also install it by installing all optional dependencies.

        ```
        python -m pip install "memo[all]"
        ```

    Arguments:
        url: web url to post json to
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            with httpx.Client() as client:
                _ = client.post(url, data={**kwargs, **result})
            return result

        return wrapper

    return decorator



import orjson
from typing import Callable, List
from functools import wraps


def memlist(data: List):
    """
    Remembers input/output of a function in python list.

    Arguments:
        data: a list to push received data into

    Example

    ```python
    from memo import memlist

    data = []

    @memlist(data=data)
    def simulate(a, b):
        return {"result": a + b}

    # The `@memlist` decorator will allow the inputs/outputs to
    # be saved in the provided `data` list.
    for a in range(5):
        for b in range(10):
            simulate(a=a, b=b)

    assert len(data) == 50

    # If we keep running more loops the list will grow.
    for a in range(6, 10 + 1):
        for b in range(11, 20 + 1):
            simulate(a=a, b=b)

    assert len(data) == 100
    ```
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            data.append({**kwargs, **result})
            return result

        return wrapper

    return decorator


def memfile(filepath: str):
    """
    Remembers input/output of a function in a jsonl file on disk.

    Arguments:
        filepath: path to write data to

    ```python
    from memo import memfile

    @memfile(filepath="tmpfile.jsonl")
    def simulate(a, b):
        return {"result": a + b}

    for a in range(5):
        for b in range(10):
            simulate(a=a, b=b)
    ```
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            with open(filepath, "a") as f:
                ser = orjson.dumps(
                    {**kwargs, **result},
                    option=orjson.OPT_NAIVE_UTC | orjson.OPT_SERIALIZE_NUMPY,
                )
                f.write(ser.decode("utf-8") + "\n")
            return result

        return wrapper

    return decorator


def memfunc(callback: Callable):
    """
    Remembers input/output of a function by printing.

    Arguments:
        callback: callback function that receives a dictionary with logged info

    ```python
    from memo import memfunc, memlist

    data = []

    @memlist(data=data)
    @memfunc(callback=print)
    def simulate(a, b):
        return {"result": a + b}

    for a in range(5):
        for b in range(10):
            simulate(a=a, b=b)

    # You should now see print statements, and this holds:
    assert len(data) == 50
    ```
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            callback({**kwargs, **result})
            return result

        return wrapper

    return decorator



from memo import grid, random_grid


def test_grid():
    inputs = [i for i in grid(a="abc", b="abcd")]
    assert len(inputs) == 12
    assert all(["a" in i for i in inputs])
    assert all(["b" in i for i in inputs])


def test_random_grid():
    inputs = [i for i in random_grid(n=10, a="abc", b="abcd")]
    assert len(inputs) == 10
    assert all(["a" in i for i in inputs])
    assert all(["b" in i for i in inputs])



import random
import itertools as it

from rich.progress import track


def grid(progbar=True, **kwargs):
    """
    Generates a grid of settings.

    Arguments:
        kwargs: the name of parameter is the key while the values represent items to iterate over

    Example

    ```python
    from memo import grid

    settings = list(grid(a=[1,2], b=[1, 2]))
    expected = [
         {'a': 1, 'b': 1},
         {'a': 1, 'b': 2},
         {'a': 2, 'b': 1},
         {'a': 2, 'b': 2}
    ]
    assert settings == expected

    def calc_sum(a, b):
        return {"c": a + b}

    for setting in settings:
        print(calc_sum(**setting))
    ```
    """
    settings = list(it.product(*[v for v in kwargs.values()]))
    if progbar:
        for s in track(settings, description="Grid..."):
            yield {k: v for k, v in zip(kwargs.keys(), s)}
    else:
        for s in settings:
            yield {k: v for k, v in zip(kwargs.keys(), s)}


def random_grid(progbar=True, n=30, **kwargs):
    """
    Generates a random grid settings.

    Arguments:
        kwargs: the name of parameter is the key while the values represent items to iterate over

    Example

    ```python
    from memo import random_grid

    settings = list(random_grid(n=30, a=[1,2], b=[1, 2]))
    assert len(settings) == 30
    ```
    """

    if progbar:
        for _ in track(range(n), description="Random Grid..."):
            yield {k: random.choice(v) for k, v in kwargs.items()}
    else:
        for _ in range(n):
            yield {k: random.choice(v) for k, v in kwargs.items()}



import pathlib

import pytest

from mktestdocs import check_md_file, check_docstring
from memo import memlist, memfunc, memfile, time_taken, grid, random_grid

files = [str(p) for p in pathlib.Path("docs").glob("*.md")] + ["README.md"]
functions = [memlist, memfunc, memfile, time_taken, grid, random_grid]


@pytest.mark.parametrize("fpath", files)
def test_files_good(fpath):
    check_md_file(fpath=fpath)


@pytest.mark.parametrize("func", functions, ids=lambda d: d.__name__)
def test_docstrings(func):
    check_docstring(func)
